cpdef double doublegaussian(double x, double mean, double sigmal, double sigmar)

cpdef double ugaussian(double x, double mean, double sigma)

cpdef double gaussian(double x, double mean, double sigma)

cpdef double crystalball(double x,double alpha,double n,double mean,double sigma)

cpdef double argus(double x, double c, double chi, double p)

cpdef double cruijff(double x, double m0, double sigma_L, double sigma_R, double alpha_L, double alpha_R)

cpdef double linear(double x, double m, double c)

cpdef double poly2(double x, double a, double b, double c)

cpdef double poly3(double x, double a, double b, double c, double d)

cpdef double novosibirsk(double x, double width, double peak, double tail)

cpdef double rtv_breitwigner(double x, double m, double gamma)

cpdef double cauchy(double x, double m, double gamma)
